// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.base;

private import std.string, std.c.stdlib;

private import dfl.winapi, dfl.drawing, dfl.event;


private extern(C) char* strcpy(char*, char*);


interface IWin32Window // docmain
{
	HWND handle(); // getter
}


class DflException: Exception // docmain
{
	this(char[] msg)
	{
		super(msg);
	}
}


class StringObject
{
	char[] value;
	
	
	this(char[] str)
	{
		this.value = str;
	}
	
	
	override char[] toString()
	{
		return value;
	}
	
	
	override int opEquals(Object o)
	{
		return value == o.toString(); // ?
	}
	
	
	int opEquals(StringObject s)
	{
		return value == s.value;
	}
	
	
	override int opCmp(Object o)
	{
		return std.string.icmp(value, o.toString()); // ?
	}
	
	
	int opCmp(StringObject s)
	{
		return std.string.icmp(value, s.value);
	}
}


enum Keys: uint // docmain
{
	NONE =     0,
	SHIFT =    0x10000,
	CONTROL =  0x20000,
	ALT =      0x40000,
	
	// Letters.
	A = 'A',
	B = 'B',
	C = 'C',
	D = 'D',
	E = 'E',
	F = 'F', 
	G = 'G',
	H = 'H',
	I = 'I',
	J = 'J',
	K = 'K',
	L = 'L',
	M = 'M',
	N = 'N',
	O = 'O',
	P = 'P',
	Q = 'Q',
	R = 'R',
	S = 'S',
	T = 'T',
	U = 'U',
	V = 'V',
	W = 'W',
	X = 'X',
	Y = 'Y',
	Z = 'Z',
	
	// Digits.
	D0 = '0',
	D1 = '1',
	D2 = '2',
	D3 = '3',
	D4 = '4',
	D5 = '5',
	D6 = '6',
	D7 = '7',
	D8 = '8',
	D9 = '9',
	
	// F - function keys.
	F1 = 112,
	F2 = 113,
	F3 = 114,
	F4 = 115,
	F5 = 116,
	F6 = 117,
	F7 = 118,
	F8 = 119,
	F9 = 120,
	F10 = 121,
	F11 = 122,
	F12 = 123,
	F13 = 124,
	F14 = 125,
	F15 = 126,
	F16 = 127,
	F17 = 128,
	F18 = 129,
	F19 = 130,
	F20 = 131,
	F21 = 132,
	F22 = 133,
	F23 = 134,
	F24 = 135,
	
	// Numbers on keypad.
	NUM_PAD0 = 96,
	NUM_PAD1 = 97,
	NUM_PAD2 = 98,
	NUM_PAD3 = 99,
	NUM_PAD4 = 100,
	NUM_PAD5 = 101,
	NUM_PAD6 = 102,
	NUM_PAD7 = 103,
	NUM_PAD8 = 104,
	NUM_PAD9 = 105,
	
	ADD = 107,
	APPS = 93, // Application.
	ATTN = 246,
	BACK = 8, // Backspace.
	CANCEL = 3,
	CAPITAL = 20, // Caps lock.
	CAPS_LOCK = 20, // Caps lock.
	CLEAR = 12,
	CONTROL_KEY = 17,
	CRSEL = 247,
	DECIMAL = 110,
	DEL = 46,
	DIVIDE = 111,
	DOWN = 40, // Down arrow.
	END = 35,
	ENTER = 13,
	ERASE_EOF = 249,
	ESCAPE = 27,
	EXECUTE = 43,
	EXSEL = 248,
	FINAL_MODE = 4, // IME final mode.
	HANGUEL_MODE = 21, // IME Hanguel key; use HANGUL_MODE.
	HANGUL_MODE = 21, // IME Hangul mode.
	HANJA_MODE = 25, // IME Hanja mode.
	HELP = 47,
	HOME = 36,
	IME_ACCEPT = 30,
	IME_CONVERT = 28,
	IME_MODE_CHANGE = 31,
	IME_NONCONVERT = 29,
	INSERT = 45,
	JUNJA_MODE = 23,
	KANA_MODE = 21,
	KANJI_MODE = 25,
	LEFT_CONTROL = 162, // Left Ctrl.
	LEFT = 37, // Left arrow.
	LINE_FEED = 10,
	LEFT_MENU = 164, // Left Alt.
	LEFT_SHIFT = 160,
	LEFT_WIN = 91, // Left Windows logo.
	MENU = 18, // Alt.
	MULTIPLY = 106,
	NEXT = 34, // Page down.
	NO_NAME = 252, // Reserved for future use.
	NUM_LOCK = 144,
	OEM8 = 223, // OEM specific.
	OEM_CLEAR = 254,
	PA1 = 253,
	PAGE_DOWN = 34,
	PAGE_UP = 33,
	PAUSE = 19,
	PLAY = 250,
	PRINT = 42,
	PRINT_SCREEN = 44,
	PROCESS_KEY = 229,
	RIGHT_CONTROL = 163, // Right Ctrl.
	RETURN = 13,
	RIGHT = 39, // Right arrow.
	RIGHT_MENU = 165, // Right Alt.
	RIGHT_SHIFT = 161,
	RIGHT_WIN = 92, // Right Windows logo.
	SCROLL = 145, // Scroll lock.
	SELECT = 41,
	SEPARATOR = 108,
	SHIFT_KEY = 16,
	SNAPSHOT = 44, // Print screen.
	SPACEBAR = 32,
	SUBTRACT = 109,
	TAB = 9,
	UP = 38, // Up arrow.
	ZOOM = 251,
	
	// Windows 2000+
	BROWSER_BACK = 166,
	BROWSER_FAVORITES = 171,
	BROWSER_FORWARD = 167,
	BROWSER_HOME = 172,
	BROWSER_REFRESH = 168,
	BROWSER_SEARCH = 170,
	BROWSER_STOP = 169,
	LAUNCH_APPLICATION1 = 182,
	LAUNCH_APPLICATION2 = 183,
	LAUNCH_MAIL = 180,
	MEDIA_NEXT_TRACK = 176,
	MEDIA_PLAY_PAUSE = 179,
	MEDIA_PREVIOUS_TRACK = 177,
	MEDIA_STOP = 178,
	OEM_BACKSLASH = 226, // OEM angle bracket or backslash.
	OEM_CLOSE_BRACKETS = 221,
	OEM_COMMA = 188,
	OEM_MINUS = 189,
	OEM_OPEN_BRACKETS = 219,
	OEM_PERIOD = 190,
	OEM_PIPE = 220,
	OEM_PLUS = 187,
	OEM_QUESTION = 191,
	OEM_QUOTES = 222,
	OEM_SEMICOLON = 186,
	OEM_TILDE = 192,
	SELECT_MEDIA = 181,
	VOLUME_DOWN = 174,
	VOLUME_MUTE = 173,
	VOLUME_UP = 175,
	
	// Bit mask to extract key code from key value.
	KEY_CODE = 0xFFFF,
	
	// Bit mask to extract modifiers from key value.
	MODIFIERS = 0xFFFF0000,
}


enum MouseButtons: uint // docmain
{
	NONE =      0,
	LEFT =      0x100000,
	RIGHT =     0x200000,
	MIDDLE =    0x400000,
	
	// Windows 2000+
	//XBUTTON1 =  0x800000,
	//XBUTTON2 =  0x1000000,
}


struct Message // docmain
{
	union
	{
		struct
		{
			HWND hWnd;
			UINT msg;
			WPARAM wParam;
			LPARAM lParam;
		}
		
		package MSG _winMsg; // .time and .pt are not always valid.
	}
	LRESULT result;
	
	
	static Message opCall(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
	{
		Message m;
		m.hWnd = hWnd;
		m.msg = msg;
		m.wParam = wParam;
		m.lParam = lParam;
		m.result = 0;
		return m;
	}
}


interface IMessageFilter // docmain
{
	// Return false to allow the message to be dispatched.
	// Filter functions cannot modify messages.
	bool preFilterMessage(inout Message m);
}


abstract class WaitHandle
{
	const int WAIT_TIMEOUT = .WAIT_TIMEOUT;
	const HANDLE INVALID_HANDLE = .INVALID_HANDLE_VALUE;
	
	
	this()
	{
		h = INVALID_HANDLE;
	}
	
	
	// Used internally.
	this(HANDLE h, bool owned = true)
	{
		this.h = h;
		this.owned = owned;
	}
	
	
	HANDLE handle() // getter
	{
		return h;
	}
	
	
	void handle(HANDLE h) // setter
	{
		this.h = h;
	}
	
	
	void close()
	{
		CloseHandle(h);
		h = INVALID_HANDLE;
	}
	
	
	~this()
	{
		if(owned)
			close();
	}
	
	
	private static DWORD _wait(WaitHandle[] handles, BOOL waitall, DWORD msTimeout)
	{
		// Some implementations fail with > 64 handles, but that will return WAIT_FAILED;
		// all implementations fail with >= 128 handles due to WAIT_ABANDONED_0 being 128.
		if(handles.length >= 128)
			goto fail;
		
		DWORD result;
		HANDLE* hs;
		//hs = new HANDLE[handles.length];
		hs = cast(HANDLE*)alloca(HANDLE.sizeof * handles.length);
		
		foreach(size_t i, WaitHandle wh; handles)
		{
			hs[i] = wh.handle;
		}
		
		result = WaitForMultipleObjects(handles.length, hs, waitall, msTimeout);
		if(WAIT_FAILED == result)
		{
			fail:
			throw new DflException("Wait failure.");
		}
		return result;
	}
	
	
	static void waitAll(WaitHandle[] handles)
	{
		return waitAll(handles, INFINITE);
	}
	
	
	static void waitAll(WaitHandle[] handles, DWORD msTimeout)
	{
		_wait(handles, true, msTimeout);
	}
	
	
	static int waitAny(WaitHandle[] handles)
	{
		return waitAny(handles, INFINITE);
	}
	
	
	static int waitAny(WaitHandle[] handles, DWORD msTimeout)
	{
		DWORD result;
		result = _wait(handles, false, msTimeout);
		return cast(int)result; // Same return info.
	}
	
	
	void waitOne()
	{
		return waitOne(INFINITE);
	}
	
	
	void waitOne(DWORD msTimeout)
	{
		DWORD result;
		result = WaitForSingleObject(handle, msTimeout);
		if(WAIT_FAILED == result)
			throw new DflException("Wait failure.");
	}
	
	
	private:
	HANDLE h;
	bit owned = true;
}


interface IAsyncResult
{
	WaitHandle asyncWaitHandle(); // getter
	
	// Usually just returns false.
	bool completedSynchronously(); // getter
	
	// When true, it is safe to release its resources.
	bool isCompleted(); // getter
}


/+
class AsyncResult: IAsyncResult
{
}
+/


interface IButtonControl // docmain
{
	DialogResult dialogResult(); // getter
	void dialogResult(DialogResult); // setter
	void notifyDefault(bool); // True if default button.
	void performClick(); // Raise click event.
}


enum DialogResult: ubyte // docmain
{
	NONE, // Modal dialog continues to run.
	
	ABORT = IDABORT,
	CANCEL = IDCANCEL,
	IGNORE = IDIGNORE,
	NO = IDNO,
	OK = IDOK,
	RETRY = IDRETRY,
	YES = IDYES,
	
	// Extra.
	CLOSE = IDCLOSE,
	HELP = IDHELP,
}


enum SortOrder: ubyte
{
	NONE,
	ASCENDING,
	DESCENDING,
}


enum View: ubyte
{
	LARGE_ICON,
	SMALL_ICON,
	LIST,
	DETAILS, // Report.
}


enum ItemBoundsPortion: ubyte
{
	ENTIRE,
	ICON,
	ITEM_ONLY, // Excludes other stuff like check boxes.
	LABEL, // Item's text.
}


enum ItemActivation: ubyte
{
	STANDARD,
	ONE_CLICK,
	TWO_CLICK,
}


enum ColumnHeaderStyle: ubyte
{
	CLICKABLE,
	NONCLICKABLE,
	NONE, // No column header.
}


enum BorderStyle: ubyte
{
	NONE,
	FIXED_3D,
	FIXED_SINGLE,
}


enum FlatStyle: ubyte
{
	STANDARD,
	FLAT,
	POPUP,
	SYSTEM,
}


enum Appearance: ubyte
{
	NORMAL,
	BUTTON,
}


enum ContentAlignment: ubyte
{
	TOP_LEFT,
	BOTTOM_CENTER,
	BOTTOM_LEFT,
	BOTTOM_RIGHT,
	MIDDLE_CENTER,
	MIDDLE_LEFT,
	MIDDLE_RIGHT,
	TOP_CENTER,
	TOP_RIGHT,
}


enum CharacterCasing: ubyte
{
	NORMAL,
	LOWER,
	UPPER,
}


// Not flags.
enum ScrollBars: ubyte
{
	NONE,
	HORIZONTAL,
	VERTICAL,
	BOTH,
}


enum HorizontalAlignment: ubyte
{
	LEFT,
	RIGHT,
	CENTER,
}


enum DrawMode: ubyte
{
	NORMAL,
	OWNER_DRAW_FIXED,
	OWNER_DRAW_VARIABLE,
}


enum DrawItemState: uint
{
	NONE = 0,
	SELECTED = 1,
	DISABLED = 2,
	CHECKED = 8,
	FOCUS = 0x10,
	DEFAULT = 0x20,
	HOT_LIGHT = 0x40,
	NO_ACCELERATOR = 0x80,
	INACTIVE = 0x100,
	NO_FOCUS_RECT = 0x200,
	COMBO_BOX_EDIT = 0x1000,
}


class LabelEditEventArgs: EventArgs
{
	this(int index)
	{
		
	}
	
	
	this(int index, char[] labelText)
	{
		this.idx = index;
		this.ltxt = labelText;
	}
	
	
	final void cancelEdit(bool byes) // setter
	{
		cancl = byes;
	}
	
	
	final bool cancelEdit() // getter
	{
		return cancl;
	}
	
	
	// The text of the label's edit.
	final char[] label() // getter
	{
		return ltxt;
	}
	
	
	// Gets the item's index.
	final int item() // getter
	{
		return idx;
	}
	
	
	private:
	int idx;
	char[] ltxt;
	bool cancl = false;
}


alias Event!(LabelEditEventArgs) LabelEditEventHandler;


class ColumnClickEventArgs: EventArgs
{
	this(int col)
	{
		this.col = col;
	}
	
	
	final int column() // getter
	{
		return col;
	}
	
	
	private:
	int col;
}


alias Event!(ColumnClickEventArgs) ColumnClickEventHandler;


class DrawItemEventArgs: EventArgs
{
	this(Graphics g, Font f, Rect r, int i, DrawItemState dis)
	{
		this(g, f, r, i , dis, Color.empty, Color.empty);
	}
	
	
	this(Graphics g, Font f, Rect r, int i, DrawItemState dis, Color fc, Color bc)
	{
		gpx = g;
		fnt = f;
		rect = r;
		idx = i;
		distate = dis;
		fcolor = fc;
		bcolor = bc;
	}
	
	
	final Color backColor() // getter
	{
		return bcolor;
	}
	
	
	final Rect bounds() // getter
	{
		return rect;
	}
	
	
	final Font font() // getter
	{
		return fnt;
	}
	
	
	final Color foreColor() // getter
	{
		return fcolor;
	}
	
	
	final Graphics graphics() // getter
	{
		return gpx;
	}
	
	
	final int index() // getter
	{
		return idx;
	}
	
	
	final DrawItemState state() // getter
	{
		return distate;
	}
	
	
	void drawBackground()
	{
		/+
		HBRUSH hbr;
		RECT _rect;
		
		hbr = bcolor.createBrush();
		try
		{
			rect.getRect(&_rect);
			FillRect(gpx.handle, &_rect, hbr);
		}
		finally
		{
			DeleteObject(hbr);
		}
		+/
		
		gpx.fillRectangle(bcolor, rect);
	}
	
	
	void drawFocusRectangle()
	{
		if(distate & DrawItemState.FOCUS)
		{
			RECT _rect;
			rect.getRect(&_rect);
			DrawFocusRect(gpx.handle, &_rect);
		}
	}
	
	
	private:
	Graphics gpx;
	Font fnt; // Suggestion; the parent's font.
	Rect rect;
	int idx;
	DrawItemState distate;
	Color fcolor, bcolor; // Suggestion; depends on item state.
}


alias Event!(DrawItemEventArgs) DrawItemEventHandler;


class MeasureItemEventArgs: EventArgs
{
	this(Graphics g, int index, int itemHeight)
	{
		gpx = g;
		idx = index;
		iheight = itemHeight;
	}
	
	
	this(Graphics g, int index)
	{
		this(g, index, 0);
	}
	
	
	final Graphics graphics() // getter
	{
		return gpx;
	}
	
	
	final int index() // getter
	{
		return idx;
	}
	
	
	final void itemHeight(int height) // setter
	{
		iheight = height;
	}
	
	
	final int itemHeight() // getter
	{
		return iheight;
	}
	
	
	final void itemWidth(int width) // setter
	{
		iwidth = width;
	}
	
	
	final int itemWidth() // getter
	{
		return iwidth;
	}
	
	
	private:
	Graphics gpx;
	int idx, iheight, iwidth = 0;
}

alias Event!(MeasureItemEventArgs) MeasureItemEventHandler;


class Cursor // docmain
{
	private static Cursor _cur;
	
	
	// Used internally.
	this(HCURSOR hcur, bool owned = true)
	{
		this.hcur = hcur;
		this.owned = owned;
	}
	
	
	~this()
	{
		if(owned)
			dispose();
	}
	
	
	void dispose()
	{
		DestroyCursor(hcur);
		hcur = null;
	}
	
	
	static void current(Cursor cur) // setter
	{
		// Keep a reference so that it doesn't get garbage collected until set again.
		_cur = cur;
		
		SetCursor(cur ? cur.hcur : null);
	}
	
	
	static Cursor current() // getter
	{
		HCURSOR hcur = GetCursor();
		return hcur ? new Cursor(hcur, false) : null;
	}
	
	
	static void clip(Rect r) // setter
	{
		RECT rect;
		r.getRect(&rect);
		ClipCursor(&rect);
	}
	
	
	static Rect clip() // getter
	{
		RECT rect;
		GetClipCursor(&rect);
		return Rect(&rect);
	}
	
	
	final HCURSOR handle() // getter
	{
		return hcur;
	}
	
	
	/+
	// TODO:
	final Size size() // getter
	{
		Size result;
		ICONINFO iinfo;
		
		if(GetIconInfo(hcur, &iinfo))
		{
			
		}
		
		return result;
	}
	+/
	
	
	// Uses the actual size.
	final void draw(Graphics g, Point pt)
	{
		DrawIconEx(g.handle, pt.x, pt.y, hcur, 0, 0, 0, null, DI_NORMAL);
	}
	
	
	/+
	// Should not stretch if bigger, but should crop if smaller.
	final void draw(Graphics g, Rect r)
	{
	}
	+/
	
	
	final void drawStretched(Graphics g, Rect r)
	{
		// DrawIconEx operates differently if the width or height is zero
		// so bail out if zero and pretend the zero size cursor was drawn.
		int width = r.width;
		if(!width)
			return;
		int height = r.height;
		if(!height)
			return;
		
		DrawIconEx(g.handle, r.x, r.y, hcur, width, height, 0, null, DI_NORMAL);
	}
	
	
	override int opEquals(Object o)
	{
		Cursor cur = cast(Cursor)o;
		if(!cur)
			return 0; // Not equal.
		return opEquals(cur);
	}
	
	
	int opEquals(Cursor cur)
	{
		return hcur == cur.hcur;
	}
	
	
	// show/hide are ref counted.
	static void hide()
	{
		ShowCursor(false);
	}
	
	
	// show/hide are ref counted.
	static void show()
	{
		ShowCursor(true);
	}
	
	
	static void position(Point pt) // setter
	{
		SetCursorPos(pt.x, pt.y);
	}
	
	
	static Point position() // getter
	{
		Point pt;
		GetCursorPos(&pt.point);
		return pt;
	}
	
	
	private:
	HCURSOR hcur;
	bool owned = true;
}


class Cursors
{
	static:
	
	Cursor appStarting() // getter
	{ return new Cursor(LoadCursorA(null, IDC_APPSTARTING), false); }
	
	Cursor arrow() // getter
	{ return new Cursor(LoadCursorA(null, IDC_ARROW), false); }
	
	Cursor cross() // getter
	{ return new Cursor(LoadCursorA(null, IDC_CROSS), false); }
	
	//Cursor default() // getter
	Cursor defaultCursor() // getter
	{ return arrow; }
	
	Cursor hand() // getter
	{
		version(SUPPORTS_HAND_CURSOR) // Windows 98+
		{
			return new Cursor(LoadCursorA(null, IDC_HAND), false);
		}
		else
		{
			static HCURSOR hcurHand = null;
			
			if(!hcurHand)
			{
				hcurHand = LoadCursorA(null, IDC_HAND);
				if(!hcurHand) // Must be Windows 95, so load the cursor from winhlp32.exe.
				{
					UINT len;
					char[MAX_PATH] winhlppath;
					
					len = GetWindowsDirectoryA(winhlppath, winhlppath.length - 16);
					if(!len || len > winhlppath.length - 16)
					{
						load_failed:
						return arrow; // Just fall back to a normal arrow.
					}
					strcpy(cast(char*)winhlppath + len, "\\winhlp32.exe");
					
					HINSTANCE hinstWinhlp;
					hinstWinhlp = LoadLibraryExA(winhlppath, null, LOAD_LIBRARY_AS_DATAFILE);
					if(!hinstWinhlp)
						goto load_failed;
					
					HCURSOR hcur;
					hcur = LoadCursorA(hinstWinhlp, cast(char*)106);
					if(!hcur) // No such cursor resource.
					{
						FreeLibrary(hinstWinhlp);
						goto load_failed;
					}
					hcurHand = CopyCursor(hcur);
					if(!hcurHand)
					{
						FreeLibrary(hinstWinhlp);
						//throw new DflException("Unable to copy cursor resource.");
						goto load_failed;
					}
					
					FreeLibrary(hinstWinhlp);
				}
			}
			
			assert(hcurHand);
			// Copy the cursor and own it here so that it's safe to dispose it.
			return new Cursor(CopyCursor(hcurHand));
		}
	}
	
	Cursor help() // getter
	{
		HCURSOR hcur;
		hcur = LoadCursorA(null, IDC_HELP);
		if(!hcur) // IDC_HELP might not be supported on Windows 95, so fall back to a normal arrow.
			return arrow;
		return new Cursor(hcur);
	}
	
	Cursor hSplit() // getter
	{
		// ...
		return sizeNS;
	}
	
	Cursor iBeam() // getter
	{ return new Cursor(LoadCursorA(null, IDC_IBEAM), false); }
	
	Cursor no() // getter
	{ return new Cursor(LoadCursorA(null, IDC_NO), false); }
	
	Cursor sizeAll() // getter
	{ return new Cursor(LoadCursorA(null, IDC_SIZEALL), false); }
	
	Cursor sizeNESW() // getter
	{ return new Cursor(LoadCursorA(null, IDC_SIZENESW), false); }
	
	Cursor sizeNS() // getter
	{ return new Cursor(LoadCursorA(null, IDC_SIZENS), false); }
	
	Cursor sizeNWSE() // getter
	{ return new Cursor(LoadCursorA(null, IDC_SIZENWSE), false); }
	
	Cursor sizeWE() // getter
	{ return new Cursor(LoadCursorA(null, IDC_SIZEWE), false); }
	
	/+
	// Insertion point.
	Cursor upArrow() // getter
	{
		// ...
	}
	+/
	
	Cursor vSplit() // getter
	{
		// ...
		return sizeWE;
	}
	
	Cursor waitCursor() // getter
	{ return new Cursor(LoadCursorA(null, IDC_WAIT), false); }
}

