// Not actually part of forms, but is needed.
// This code is public domain.

module dfl.event;


// Create an event handler.
struct Event(TArgs : EventArgs = EventArgs) // docmain
{
	alias void delegate(Object, Args) Handler;
	alias TArgs Args;
	
	
	void addHandler(Handler handler)
	{
		array ~= handler;
	}
	
	
	void opCatAssign(Handler handler)
	{
		array ~= handler;
	}
	
	
	void removeHandler(Handler handler)
	{
		uint i;
		for(i = 0; i != array.length; i++)
		{
			if(handler == array[i])
			{
				if(!i)
					array = array[1 .. array.length];
				else if(i == array.length - 1)
					array = array[0 .. i];
				else
					array = array[0 .. i] ~ array[i + 1 .. array.length];
				break;
			}
		}
	}
	
	
	void opCall(Object sender, Args args)
	{
		// Keep a local reference so that handlers
		// may be added and removed during iteration.
		Handler[] local = array;
		
		foreach(Handler handler; local)
		{
			handler(sender, args);
		}
	}
	
	
	Handler[] handlers() // getter
	{
		return array;
	}
	
	
	private:
	Handler[] array;
}


class EventArgs // docmain
{
	/+
	private static byte[] buf;
	private import std.gc;
	
	
	new(uint sz)
	{
		void* result;
		
		// synchronized // Slows it down a lot.
		{
			if(sz > buf.length)
				buf = new byte[100 + sz];
			
			result = buf[0 .. sz];
			buf = buf[sz .. buf.length];
		}
		
		// std.gc.addRange(result, result + sz); // So that it can contain pointers.
		return result;
	}
	+/
	
	
	/+
	delete(void* p)
	{
		std.gc.removeRange(p);
	}
	+/
	
	
	private static EventArgs _e;
	
	
	static this()
	{
		_e = new EventArgs;
	}
	
	
	static EventArgs empty() // getter
	{
		return _e;
	}
}


// Simple event handler.
alias Event!() EventHandler;


class ThreadExceptionEventArgs: EventArgs
{
	// The exception that occured.
	this(Object e)
	{
		except = e;
	}
	
	
	final Object exception() // getter
	{
		return except;
	}
	
	
	private:
	Object except;
}


alias Event!(ThreadExceptionEventArgs) ThreadExceptionEventHandler;

