// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.groupbox;

private import dfl.control, dfl.base, dfl.button, dfl.drawing;
private import dfl.winapi, dfl.application, dfl.event;


private extern(Windows) void _initButton();


class GroupBox: ControlSuperClass // docmain
{
	override Rect displayRectangle()
	{
		// Should only calculate this upon setting the text ?
		
		int xw = GetSystemMetrics(SM_CXFRAME);
		int yw = GetSystemMetrics(SM_CYFRAME);
		const int textHeight = 13; // Hack.
		return Rect(xw, yw + textHeight, clientSize.width - xw * 2, clientSize.height - yw - textHeight - yw);
	}
	
	
	override Size defaultSize() // getter
	{
		return Size(200, 100);
	}
	
	
	version(NO_DRAG_DROP) {} else
	{
		alias Control.allowDrop allowDrop; // Override.
		
		void allowDrop(bool dyes) // setter
		{
			if(dyes)
				throw new DflException("Cannot drop on a group box.");
		}
	}
	
	
	this()
	{
		_initButton();
		
		wstyle |= BS_GROUPBOX /+ | WS_TABSTOP +/; // Should WS_TABSTOP be set?
		wexstyle |= WS_EX_CONTROLPARENT;
		wclassStyle = buttonClassStyle;
	}
	
	
	/+
	override void createHandle()
	{
		if(isHandleCreated)
			return;
		
		createClassHandle(BUTTON_CLASSNAME);
		
		onHandleCreated(EventArgs.empty);
	}
	+/
	
	
	protected override void createParams(inout CreateParams cp)
	{
		super.createParams(cp);
		
		cp.className = BUTTON_CLASSNAME;
	}
	
	
	protected override void wndProc(inout Message msg)
	{
		switch(msg.msg)
		{
			case WM_NCHITTEST:
				Control.defWndProc(msg);
				break;
			
			default:
				super.wndProc(msg);
		}
	}
	
	
	protected override void onPaintBackground(PaintEventArgs ea)
	{
		//Control.onPaintBackground(ea); // DMD 0.106: not accessible.
		
		RECT rect;
		ea.clipRectangle.getRect(&rect);
		FillRect(ea.graphics.handle, &rect, hbrBg);
	}
	
	
	protected override void prevWndProc(inout Message msg)
	{
		msg.result = CallWindowProcA(buttonPrevWndProc, msg.hWnd, msg.msg, msg.wParam, msg.lParam);
	}
}

