// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.label;

private import dfl.base, dfl.control, dfl.winapi, dfl.application;
private import dfl.event;


private extern(Windows) void _initLabel();


class Label: ControlSuperClass // docmain
{
	this()
	{
		_initLabel();
		
		wclassStyle = labelClassStyle;
	}
	
	
	void borderStyle(BorderStyle bs) // setter
	{
		switch(bs)
		{
			case BorderStyle.FIXED_3D:
				_style(_style() & ~WS_BORDER);
				_exStyle(_exStyle() | WS_EX_CLIENTEDGE);
				break;
				
			case BorderStyle.FIXED_SINGLE:
				_exStyle(_exStyle() & ~WS_EX_CLIENTEDGE);
				_style(_style() | WS_BORDER);
				break;
				
			case BorderStyle.NONE:
				_style(_style() & ~WS_BORDER);
				_exStyle(_exStyle() & ~WS_EX_CLIENTEDGE);
				break;
		}
		
		if(created)
		{
			redrawEntire();
		}
	}
	
	
	BorderStyle borderStyle() // getter
	{
		if(_exStyle() & WS_EX_CLIENTEDGE)
			return BorderStyle.FIXED_3D;
		else if(_style() & WS_BORDER)
			return BorderStyle.FIXED_SINGLE;
		return BorderStyle.NONE;
	}
	
	
	final void useMnemonic(bool byes) // setter
	{
		if(byes)
			_style(_style() & ~SS_NOPREFIX);
		else
			_style(_style() | SS_NOPREFIX);
		
		if(created)
			invalidate();
	}
	
	
	final bool useMnemonic() // getter
	{
		return (_style() & SS_NOPREFIX) == 0;
	}
	
	
	/+
	void textAlign(ContentAlignment calign) // setter
	{
		LONG wl = _style() & ~(SS_BOTTOM | SS_CENTER | SS_TOP | SS_RIGHT | SS_LEFT | SS_VCENTER);
		
		switch(calign)
		{
			case ContentAlignment.TOP_LEFT:
				wl |= SS_TOP | SS_LEFT;
				break;
			
			case ContentAlignment.BOTTOM_CENTER:
				wl |= SS_BOTTOM | SS_CENTER;
				break;
			
			case ContentAlignment.BOTTOM_LEFT:
				wl |= SS_BOTTOM | SS_LEFT;
				break;
			
			case ContentAlignment.BOTTOM_RIGHT:
				wl |= SS_BOTTOM | SS_RIGHT;
				break;
			
			case ContentAlignment.MIDDLE_CENTER:
				wl |= SS_CENTER | SS_VCENTER;
				break;
			
			case ContentAlignment.MIDDLE_LEFT:
				wl |= SS_VCENTER | SS_LEFT;
				break;
			
			case ContentAlignment.MIDDLE_RIGHT:
				wl |= SS_VCENTER | SS_RIGHT;
				break;
			
			case ContentAlignment.TOP_CENTER:
				wl |= SS_TOP | SS_CENTER;
				break;
			
			case ContentAlignment.TOP_RIGHT:
				wl |= SS_TOP | SS_RIGHT;
				break;
		}
		
		_style(wl);
		invalidate(); // ?
	}
	
	
	ContentAlignment textAlign() // getter
	{
		LONG wl = _style();
		
		if(wl & SS_VCENTER) // Middle.
		{
			if(wl & SS_CENTER)
				return ContentAlignment.MIDDLE_CENTER;
			if(wl & SS_RIGHT)
				return ContentAlignment.MIDDLE_RIGHT;
			return ContentAlignment.MIDDLE_LEFT;
		}
		else if(wl & SS_BOTTOM) // Bottom.
		{
			if(wl & SS_CENTER)
				return ContentAlignment.BOTTOM_CENTER;
			if(wl & SS_RIGHT)
				return ContentAlignment.BOTTOM_RIGHT;
			return ContentAlignment.BOTTOM_LEFT;
		}
		else // Top.
		{
			if(wl & SS_CENTER)
				return ContentAlignment.TOP_CENTER;
			if(wl & SS_RIGHT)
				return ContentAlignment.TOP_RIGHT;
			return ContentAlignment.TOP_LEFT;
		}
	}
	+/
	
	
	/+
	override void createHandle()
	{
		if(isHandleCreated)
			return;
		
		createClassHandle(LABEL_CLASSNAME);
		
		onHandleCreated(EventArgs.empty);
	}
	+/
	
	
	protected override void createParams(inout CreateParams cp)
	{
		super.createParams(cp);
		
		cp.className = LABEL_CLASSNAME;
	}
	
	
	protected override void wndProc(inout Message msg)
	{
		switch(msg.msg)
		{
			case WM_NCHITTEST:
				Control.defWndProc(msg);
				break;
			
			default:
				super.wndProc(msg);
		}
	}
	
	
	protected override void prevWndProc(inout Message msg)
	{
		msg.result = CallWindowProcA(labelPrevWndProc, msg.hWnd, msg.msg, msg.wParam, msg.lParam);
	}
}

