// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.messagebox;

private import std.string;

private import dfl.winapi, dfl.base;


enum MsgBoxButtons
{
	ABORT_RETRY_IGNORE = MB_ABORTRETRYIGNORE,
	OK = MB_OK,
	OK_CANCEL = MB_OKCANCEL,
	RETRY_CANCEL = MB_RETRYCANCEL,
	YES_NO = MB_YESNO,
	YES_NO_CANCEL = MB_YESNOCANCEL,
}


enum MsgBoxIcon
{
	NONE = 0,
	
	ASTERISK = MB_ICONASTERISK,
	ERROR = MB_ICONERROR,
	EXCLAMATION = MB_ICONEXCLAMATION,
	HAND = MB_ICONHAND,
	INFORMATION = MB_ICONINFORMATION,
	QUESTION = MB_ICONQUESTION,
	STOP = MB_ICONSTOP,
	WARNING = MB_ICONWARNING,
}


enum MsgBoxDefaultButton
{
	BUTTON1 = MB_DEFBUTTON1,
	BUTTON2 = MB_DEFBUTTON2,
	BUTTON3 = MB_DEFBUTTON3,
	
	// Extra.
	BUTTON4 = MB_DEFBUTTON4,
}


enum MsgBoxOptions
{
	DEFAULT_DESKTOP_ONLY = MB_DEFAULT_DESKTOP_ONLY,
	RIGHT_ALIGN = MB_RIGHT,
	LEFT_ALIGN = MB_RTLREADING,
	SERVICE_NOTIFICATION = MB_SERVICE_NOTIFICATION,
}


DialogResult msgBox(char[] txt) // docmain
{
	return cast(DialogResult)dfl.utf.messageBox(GetActiveWindow(), txt, \0, MB_OK);
}


DialogResult msgBox(IWin32Window owner, char[] txt) // docmain
{
	return cast(DialogResult)dfl.utf.messageBox(owner ? owner.handle : GetActiveWindow(),
		txt, \0, MB_OK);
}


DialogResult msgBox(char[] txt, char[] caption) // docmain
{
	return cast(DialogResult)dfl.utf.messageBox(GetActiveWindow(), txt, caption, MB_OK);
}


DialogResult msgBox(IWin32Window owner, char[] txt, char[] caption) // docmain
{
	return cast(DialogResult)dfl.utf.messageBox(owner ? owner.handle : GetActiveWindow(),
		txt, caption, MB_OK);
}


DialogResult msgBox(char[] txt, char[] caption, MsgBoxButtons buttons) // docmain
{
	return cast(DialogResult)dfl.utf.messageBox(GetActiveWindow(), txt, caption, buttons);
}


DialogResult msgBox(IWin32Window owner, char[] txt, char[] caption,
	MsgBoxButtons buttons) // docmain
{
	return cast(DialogResult)dfl.utf.messageBox(owner ? owner.handle : GetActiveWindow(),
		txt, caption, buttons);
}


DialogResult msgBox(char[] txt, char[] caption, MsgBoxButtons buttons,
	MsgBoxIcon icon) // docmain
{
	return cast(DialogResult)dfl.utf.messageBox(GetActiveWindow(), txt,
		caption, buttons | icon);
}


DialogResult msgBox(IWin32Window owner, char[] txt, char[] caption, MsgBoxButtons buttons,
	MsgBoxIcon icon) // docmain
{
	return cast(DialogResult)dfl.utf.messageBox(owner ? owner.handle : GetActiveWindow(),
		txt, caption, buttons | icon);
}


DialogResult msgBox(char[] txt, char[] caption, MsgBoxButtons buttons, MsgBoxIcon icon,
	MsgBoxDefaultButton defaultButton) // docmain
{
	return cast(DialogResult)dfl.utf.messageBox(GetActiveWindow(), txt,
		caption, buttons | icon | defaultButton);
}


DialogResult msgBox(IWin32Window owner, char[] txt, char[] caption, MsgBoxButtons buttons,
	MsgBoxIcon icon, MsgBoxDefaultButton defaultButton) // docmain
{
	return cast(DialogResult)dfl.utf.messageBox(owner ? owner.handle : GetActiveWindow(),
		txt, caption, buttons | icon | defaultButton);
}


DialogResult msgBox(IWin32Window owner, char[] txt, char[] caption, MsgBoxButtons buttons,
	MsgBoxIcon icon, MsgBoxDefaultButton defaultButton, MsgBoxOptions options) // docmain
{
	return cast(DialogResult)dfl.utf.messageBox(owner ? owner.handle : GetActiveWindow(),
		txt, caption, buttons | icon | defaultButton | options);
}


deprecated class MessageBox
{
	private this() {}
	
	
	static:
	deprecated alias msgBox show;
}


deprecated alias MessageBox.show messageBox;

deprecated alias MsgBoxOptions MessageBoxOptions;
deprecated alias MsgBoxDefaultButton MessageBoxDefaultButton;
deprecated alias MsgBoxButtons MessageBoxButtons;
deprecated alias MsgBoxIcon MessageBoxIcon;

