// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.splitter;

private import dfl.control, dfl.winapi, dfl.base, dfl.drawing;
private import dfl.event;


class SplitterEventArgs: EventArgs
{
	this(int x, int y, int splitX, int splitY)
	{
		_x = x;
		_y = y;
		_splitX = splitX;
		_splitY = splitY;
	}
	
	
	final int x() // getter
	{
		return _x;
	}
	
	
	final int y() // getter
	{
		return _y;
	}
	
	
	final void splitX(int val) // setter
	{
		_splitX = val;
	}
	
	
	final int splitX() // getter
	{
		return _splitX;
	}
	
	
	final void splitY(int val) // setter
	{
		_splitY = val;
	}
	
	
	final int splitY() // getter
	{
		return _splitY;
	}
	
	
	private:
	int _x, _y, _splitX, _splitY;
}


alias Event!(SplitterEventArgs) SplitterEventHandler;


class Splitter: Control // docmain
{
	this()
	{
		// DMD 0.95: need 'this' to access member dock
		this.dock = DockStyle.LEFT;
	}
	
	
	alias Control.anchor anchor; // Overload.
	
	override void anchor(AnchorStyles a) // setter
	{
		throw new DflException("Splitter cannot be anchored.");
	}
	
	
	alias Control.dock dock; // Overload.
	
	void dock(DockStyle ds) // setter
	{
		switch(ds)
		{
			case DockStyle.LEFT:
			case DockStyle.RIGHT:
				//cursor = new Cursor(LoadCursorA(null, IDC_SIZEWE), false);
				cursor = Cursors.vSplit;
				break;
			
			case DockStyle.TOP:
			case DockStyle.BOTTOM:
				//cursor = new Cursor(LoadCursorA(null, IDC_SIZENS), false);
				cursor = Cursors.hSplit;
				break;
			
			default:
				throw new DflException("Invalid splitter dock.");
		}
		
		super.dock(ds);
	}
	
	
	protected override void onMouseDown(MouseEventArgs mea)
	{
		super.onMouseDown(mea);
		
		if(mea.button == MouseButtons.LEFT)
		{
			capture = true;
			
			downing = true;
			downpos = Point(mea.x, mea.y);
		}
	}
	
	
	protected override void onMouseMove(MouseEventArgs mea)
	{
		super.onMouseMove(mea);
		
		if(downing)
			onSplitterMoving(new SplitterEventArgs(mea.x(), mea.y(), left(), top()));
	}
	
	
	// TODO: check if correct implementation.
	protected override void onMove(EventArgs ea)
	{
		super.onMove(ea);
		
		Point curpos = Cursor.position;
		curpos = pointToClient(curpos);
		onSplitterMoved(new SplitterEventArgs(curpos.x, curpos.y, left(), top()));
	}
	
	
	protected override void onMouseUp(MouseEventArgs mea)
	{
		super.onMouseUp(mea);
		
		if(downing)
		{
			capture = false;
			
			downing = false;
			
			/+
			if(leftright)
				left = left + mea.x;
			else
				top = left + mea.y;
			+/
			
			Control splat; // Splitted.
			int adj, val, vx;
			
			// DMD 0.95: need 'this' to access member dock
			//switch(dock())
			switch(this.dock())
			{
				case DockStyle.LEFT:
					foreach(Control ctrl; parent.controls())
					{
						// DMD 0.95: overloads int(Object o) and int(Control ctrl) both match argument list for opEquals
						//if(ctrl == this)
						if(ctrl == cast(Control)this)
						{
							if(splat)
							{
								val = left - splat.left + mea.x - downpos.x;
								if(val < msize)
									val = msize;
								splat.width = val;
							}
							break;
						}
						splat = ctrl;
					}
					break;
				
				case DockStyle.RIGHT:
					foreach(Control ctrl; parent.controls())
					{
						// DMD 0.95: overloads int(Object o) and int(Control ctrl) both match argument list for opEquals
						//if(ctrl == this)
						if(ctrl == cast(Control)this)
						{
							if(splat)
							{
								adj = right - splat.left + mea.x - downpos.x;
								val = splat.width - adj;
								vx = splat.left + adj;
								if(val < msize)
								{
									vx -= msize - val;
									val = msize;
								}
								splat.bounds = Rect(vx, splat.top, val, splat.height);
							}
							break;
						}
						splat = ctrl;
					}
					break;
				
				case DockStyle.TOP:
					foreach(Control ctrl; parent.controls())
					{
						// DMD 0.95: overloads int(Object o) and int(Control ctrl) both match argument list for opEquals
						//if(ctrl == this)
						if(ctrl == cast(Control)this)
						{
							if(splat)
							{
								val = top - splat.top + mea.y - downpos.y;
								if(val < msize)
									val = msize;
								splat.height = val;
							}
							break;
						}
						splat = ctrl;
					}
					break;
				
				case DockStyle.BOTTOM:
					foreach(Control ctrl; parent.controls())
					{
						// DMD 0.95: overloads int(Object o) and int(Control ctrl) both match argument list for opEquals
						//if(ctrl == this)
						if(ctrl == cast(Control)this)
						{
							if(splat)
							{
								adj = bottom - splat.top + mea.y - downpos.y;
								val = splat.height - adj;
								vx = splat.top + adj;
								if(val < msize)
								{
									vx -= msize - val;
									val = msize;
								}
								splat.bounds = Rect(splat.left, vx, splat.width, val);
							}
							break;
						}
						splat = ctrl;
					}
					break;
			}
			
			// This is needed when the moved control first overlaps the splitter and the splitter
			// gets bumped over, causing a little area to not be updated correctly.
			// I'll fix it someday.
			parent.invalidate(true);
			
			// Event..
		}
	}
	
	
	/+
	// Not quite sure how to implement this yet.
	// Might need to scan all controls until one of:
	//    Control with opposite dock (right if left dock): stay -mextra- away from it,
	//    Control with fill dock: that control can't have less than -mextra- width,
	//    Reached end of child controls: stay -mextra- away from the edge.
	
	final void minExtra(int min) // setter
	{
		mextra = min;
	}
	
	
	final int minExtra() // getter
	{
		return mextra;
	}
	+/
	
	
	final void minSize(int min) // setter
	{
		msize = min;
	}
	
	
	final int minSize() // getter
	{
		return msize;
	}
	
	
	/+
	// TODO: implement.
	
	final void splitPosition(int pos) // setter
	{
		
	}
	
	
	// -1 if not docked to a control.
	final int splitPosition() // getter
	{
		
	}
	+/
	
	
	SplitterEventHandler splitterMoved;
	SplitterEventHandler splitterMoving;
	
	
	protected:
	override Size defaultSize() // getter
	{
		return Size(GetSystemMetrics(SM_CXSIZEFRAME), GetSystemMetrics(SM_CYSIZEFRAME));
	}
	
	
	void onSplitterMoving(SplitterEventArgs sea)
	{
		splitterMoving(this, sea);
	}
	
	
	void onSplitterMoved(SplitterEventArgs sea)
	{
		splitterMoving(this, sea);
	}
	
	
	private:
	bit downing = false;
	Point downpos;
	int msize = 25; // Min size of control that's being sized from the splitter.
	int mextra = 25; // Min size of the control on the opposite side.
}

