// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.tabcontrol;

private import std.string;

private import dfl.control, dfl.panel, dfl.winapi, dfl.drawing;
private import dfl.application, dfl.event, dfl.base, dfl.collections;


private extern(Windows) void _initTabcontrol();


class TabPage: Panel
{
	this(char[] tabText)
	{
		this();
		
		this.text = tabText;
	}
	
	
	this()
	{
		Application.ppin(cast(void*)this);
		
		wstyle &= ~WS_VISIBLE;
	}
	
	
	// imageIndex
	
	
	alias Panel.text text; // Overload with Panel.text.
	
	override void text(char[] newText) // setter
	{
		// Note: this probably causes toStringz() to be called twice,
		// allocating 2 of the same string.
		
		super.text = newText;
		
		if(created)
		{
			TabControl tc;
			tc = cast(TabControl)parent;
			if(tc)
				tc.updateTabText(this, newText);
		}
	}
	
	
	/+
	final void toolTipText(char[] ttt) // setter
	{
		// TODO: ...
	}
	
	
	final char[] toolTipText() // getter
	{
		// TODO: ...
		return null;
	}
	+/
	
	
	/+ package +/ protected int _rtype() // package
	{
		return 4;
	}
	
	
	protected override void setBoundsCore(int x, int y, int width, int height, BoundsSpecified specified)
	{
		assert(0); // Cannot set bounds of TabPage; it is done automatically.
	}
	
	
	package final void realBounds(Rect r) // setter
	{
		// DMD 0.124: if I don't put this here, super.setBoundsCore ends up calling setBoundsCore instead of super.setBoundsCore.
		void delegate(int, int, int, int, BoundsSpecified) _foo = &setBoundsCore;
		
		super.setBoundsCore(r.x, r.y, r.width, r.height, BoundsSpecified.ALL);
	}
	
	
	protected override void setVisibleCore(bool byes)
	{
		assert(0); // Cannot set visibility of TabPage; it is done automatically.
	}
	
	
	package final void realVisible(bool byes) // setter
	{
		// DMD 0.124: if I don't put this here, super.setVisibleCore ends up calling setVisibleCore instead of super.setVisibleCore.
		void delegate(bool byes) _foo = &setVisibleCore;
		
		super.setVisibleCore(byes);
	}
}


class TabPageCollection //: IList
{
	this(TabControl owner)
	in
	{
		assert(owner.tchildren is null);
	}
	body
	{
		tc = owner;
	}
	
	
	bool isFixedSize() // getter
	{
		return false;
	}
	
	
	bool isReadOnly() // getter
	{
		return false;
	}
	
	
	bool isSynchronized() // getter
	{
		return false;
	}
	
	
	Object syncRoot() // getter
	{
		return this;
	}
	
	
	// Shallow copy.
	Array clone()
	{
		Array result;
		result = new Array(_pages.length);
		Array.copy(_pages, result, _pages.length);
		return result;
	}
	
	
	// D-style.
	final Array dup() // getter
	{
		return clone();
	}
	
	
	void copyTo(Array dest, int destIndex)
	{
		Array.copy(_pages, 0, dest, destIndex, _pages.length);
	}
	
	
	void add(Object o)
	{
		TabPage page;
		page = cast(TabPage)o;
		if(page)
		{
			return add(page);
		}
		else
		{
			return add(o.toString()); // ?
		}
	}
	
	
	private void verifyNoParent(TabPage page)
	{
		if(page.parent)
			throw new DflException("TabPage already has a parent.");
	}
	
	
	package final bool created() // getter
	{
		return tc && tc.created();
	}
	
	
	void add(TabPage page)
	{
		return insert(cast(int)_pages.length, page);
	}
	
	
	void add(char[] text)
	{
		return add(new TabPage(text));
	}
	
	
	void addRange(Object[] range)
	{
		foreach(Object o; range)
		{
			add(o);
		}
	}
	
	
	void addRange(TabPage[] range)
	{
		foreach(TabPage page; range)
		{
			add(page);
		}
	}
	
	
	void addRange(char[][] range)
	{
		foreach(char[] s; range)
		{
			add(s);
		}
	}
	
	
	void clear()
	{
		TabPage[] pages = _pages;
		_pages = null;
		
		if(created)
		{
			Message m;
			m = Message(tc.handle, TCM_DELETEALLITEMS, 0, 0);
			tc.prevWndProc(m);
		}
		
		foreach(TabPage page; pages)
		{
			page.parent = null;
		}
	}
	
	
	bool contains(Object value)
	{
		return indexOf(value) != -1;
	}
	
	
	bool contains(TabPage value)
	{
		return indexOf(value) != -1;
	}
	
	
	int indexOf(Object value)
	{
		TabPage page;
		page = cast(TabPage)page;
		if(page)
		{
			return indexOf(page);
		}
		else
		{
			foreach(int i, TabPage page; _pages)
			{
				if(value == page)
					return i;
			}
			return -1;
		}
	}
	
	
	int indexOf(TabPage value)
	{
		int result = -1;
		foreach(int i, TabPage page; _pages)
		{
			//if(value == page)
			{
				if(value is page)
				{
					result = i;
					break;
				}
				
				//if(result == -1)
				//	result = i;
			}
		}
		return result;
	}
	
	
	void insert(int index, Object value)
	{
		TabPage page;
		page = cast(TabPage)value;
		if(page)
		{
			insert(index, page);
		}
		else
		{
			insert(index, value.toString()); // ?
		}
	}
	
	
	void insert(int index, char[] value)
	{
		insert(index, new TabPage(value));
	}
	
	
	void insert(int i, TabPage page)
	{
		verifyNoParent(page);
		
		//page.visible = false;
		assert(page.visible == false);
		assert(!(tc is null));
		page.parent = tc;
		
		if(i < 0)
			i = 0;
		/+if(i <= 0)
			_pages = (&page)[0 .. 1].dup ~ _pages; // DMD 0.124: .dup needed for some reason.
		else+/ if(i >= cast(int)_pages.length)
			_pages ~= page;
		else
			_pages = _pages[0 .. i] ~ (&page)[0 .. 1] ~ _pages[i .. _pages.length];
		
		if(created)
		{
			TC_ITEMA tci;
			// TODO: TCIF_RTLREADING flag based on rightToLeft property.
			tci.mask = TCIF_TEXT | TCIF_PARAM;
			tci.pszText = toStringz(page.text);
			static assert(tci.lParam.sizeof >= (void*).sizeof);
			tci.lParam = cast(LPARAM)cast(void*)page;
			
			Message m;
			m = Message(tc.handle, TCM_INSERTITEMA, i, cast(LPARAM)&tci);
			tc.prevWndProc(m);
			assert(cast(int)m.result != -1);
		}
		
		//return i;
	}
	
	
	void remove(Object value)
	{
		int i;
		i = indexOf(value);
		if(i != -1)
			removeAt(i);
	}
	
	
	void remove(TabPage page)
	{
		int i;
		i = indexOf(page);
		if(i != -1)
			removeAt(i);
	}
	
	
	void removeAt(int index)
	{
		TabPage page;
		page = _pages[index];
		
		if(!index)
			_pages = _pages[1 .. _pages.length];
		else if(index == _pages.length - 1)
			_pages = _pages[0 .. index];
		else if(index > 0 && index < cast(int)_pages.length)
			_pages = _pages[0 .. index] ~ _pages[index + 1 .. _pages.length];
		
		if(created)
		{
			Message m;
			m = Message(tc.handle, TCM_DELETEITEM, index, 0);
			tc.prevWndProc(m);
		}
	}
	
	
	int count() // getter
	{
		return _pages.length;
	}
	
	
	void opIndexAssign(Object value, int index) // setter
	{
		//array[index] = value;
		
		removeAt(index);
		//add(value);
		insert(index, value);
	}
	
	
	// Covariant return with interface IList.
	TabPage opIndex(int index) // getter
	{
		return _pages[index];
	}
	
	
	IEnumerator getEnumerator()
	{
		return new PageEnum(_pages);
	}
	
	
	private:
	TabControl tc;
	TabPage[] _pages = null;
	
	
	void doPages()
	in
	{
		assert(created);
	}
	body
	{
		Rect area;
		area = tc.displayRectangle;
		
		Message m;
		m.hWnd = tc.handle;
		m.msg = TCM_INSERTITEMA;
		
		foreach(int i, TabPage page; _pages)
		{
			TC_ITEMA tci;
			// TODO: TCIF_RTLREADING flag based on rightToLeft property.
			tci.mask = TCIF_TEXT | TCIF_PARAM;
			tci.pszText = toStringz(page.text);
			static assert(tci.lParam.sizeof >= (void*).sizeof);
			tci.lParam = cast(LPARAM)cast(void*)page;
			
			m.wParam = i;
			m.lParam = cast(LPARAM)&tci;
			tc.prevWndProc(m);
			assert(cast(int)m.result != -1);
		}
	}
	
	
	static class PageEnum: IEnumerator
	{
		mixin EnumeratorForeach;
		//mixin EnumeratorForeach!(TabPage);
		
		
		this(TabPage[] pages)
		{
			this.pages = pages;
		}
		
		
		// Covariant return with IEnumerator.
		//TabPage current() // getter
		Object current() // getter
		{
			return pages[idx];
		}
		
		
		bool moveNext()
		{
			idx++;
			if(idx >= pages.length) // Note: signed-unsigned comparison.
			{
				idx = pages.length; // Prevent overflow.
				return false;
			}
			return true;
		}
		
		
		void reset()
		{
			idx = -1;
		}
		
		
		private:
		TabPage[] pages;
		int idx = -1;
	}
}


enum TabAlignment: ubyte
{
	TOP,
	BOTTOM,
	LEFT,
	RIGHT,
}


enum TabAppearance: ubyte
{
	NORMAL,
	BUTTONS,
	FLAT_BUTTONS,
}


enum TabDrawMode: ubyte
{
	NORMAL,
	OWNER_DRAW_FIXED,
}


class TabControl: ControlSuperClass // docmain
{
	this()
	{
		_initTabcontrol();
		
		wstyle |= WS_TABSTOP;
		//wexstyle |= WS_EX_CONTROLPARENT; // Doesn't let you tab to the tabs.
		ctrlStyle |= ControlStyles.SELECTABLE; // Not considered a container control even though it contains TabPage`s.
		wclassStyle = tabcontrolClassStyle;
		
		tchildren = new TabPageCollection(this);
	}
	
	
	final void alignment(TabAlignment talign) // setter
	{
		switch(talign)
		{
			case TabAlignment.TOP:
				_style(wstyle & ~(TCS_VERTICAL | TCS_RIGHT | TCS_BOTTOM));
				break;
			
			case TabAlignment.BOTTOM:
				_style((wstyle & ~(TCS_VERTICAL | TCS_RIGHT)) | TCS_BOTTOM);
				break;
			
			case TabAlignment.LEFT:
				_style((wstyle & ~(TCS_BOTTOM | TCS_RIGHT)) | TCS_VERTICAL);
				break;
			
			case TabAlignment.RIGHT:
				_style((wstyle & ~TCS_BOTTOM) | TCS_VERTICAL | TCS_RIGHT);
				break;
			
			default:
				assert(0);
		}
		
		// Display rectangle changed.
		
		if(created && visible)
		{
			invalidate(true); // Update children too ?
			
			TabPage page;
			page = selectedTab;
			if(page)
				page.realBounds = displayRectangle;
		}
	}
	
	
	final TabAlignment alignment() // getter
	{
		// Note: TCS_RIGHT and TCS_BOTTOM are the same flag.
		
		if(wstyle & TCS_VERTICAL)
		{
			if(wstyle & TCS_RIGHT)
				return TabAlignment.RIGHT;
			return TabAlignment.LEFT;
		}
		else
		{
			if(wstyle & TCS_BOTTOM)
				return TabAlignment.BOTTOM;
			return TabAlignment.TOP;
		}
	}
	
	
	final void appearance(TabAppearance tappear) // setter
	{
		switch(tappear)
		{
			case TabAppearance.NORMAL:
				_style(wstyle & ~(TCS_BUTTONS | TCS_FLATBUTTONS));
				break;
			
			case TabAppearance.BUTTONS:
				_style((wstyle & ~TCS_FLATBUTTONS) | TCS_BUTTONS);
				break;
			
			case TabAppearance.FLAT_BUTTONS:
				_style(wstyle | TCS_BUTTONS | TCS_FLATBUTTONS);
				break;
			
			default:
				assert(0);
		}
		
		if(created && visible)
		{
			invalidate(false);
			
			TabPage page;
			page = selectedTab;
			if(page)
				page.realBounds = displayRectangle;
		}
	}
	
	
	final TabAppearance appearance() // getter
	{
		if(wstyle & TCS_FLATBUTTONS)
			return TabAppearance.FLAT_BUTTONS;
		if(wstyle & TCS_BUTTONS)
			return TabAppearance.BUTTONS;
		return TabAppearance.NORMAL;
	}
	
	
	final void drawMode(TabDrawMode dm) // setter
	{
		switch(dm)
		{
			case TabDrawMode.OWNER_DRAW_FIXED:
				_style(wstyle | TCS_OWNERDRAWFIXED);
				break;
			
			case TabDrawMode.NORMAL:
				_style(wstyle & ~TCS_OWNERDRAWFIXED);
				break;
			
			default:
				assert(0);
		}
	}
	
	
	final TabDrawMode drawMode() // getter
	{
		if(wstyle & TCS_OWNERDRAWFIXED)
			return TabDrawMode.OWNER_DRAW_FIXED;
		return TabDrawMode.NORMAL;
	}
	
	
	final void padding(Point pad) // setter
	{
		if(created)
		{
			SendMessageA(hwnd, TCM_SETPADDING, 0, MAKELPARAM(pad.x, pad.y));
			
			TabPage page;
			page = selectedTab;
			if(page)
				page.realBounds = displayRectangle;
		}
		
		_pad = pad;
	}
	
	
	final Point padding() // getter
	{
		return _pad;
	}
	
	
	final TabPageCollection tabPages() // getter
	{
		return tchildren;
	}
	
	
	override Rect displayRectangle() // getter
	{
		if(!created)
		{
			return super.displayRectangle(); // Hack?
		}
		else
		{
			RECT drr;
			Message m;
			drr.left = 0;
			drr.top = 0;
			drr.right = clientSize.width;
			drr.bottom = clientSize.height;
			m = Message(hwnd, TCM_ADJUSTRECT, FALSE, cast(LPARAM)&drr);
			prevWndProc(m);
			return Rect(&drr);
		}
	}
	
	
	protected override Size defaultSize() // getter
	{
		return Size(200, 200); // ?
	}
	
	
	final void multiline(bool byes) // setter
	{
		if(byes)
			_style(_style() | TCS_MULTILINE);
		else
			_style(_style() & ~TCS_MULTILINE);
		
		TabPage page;
		page = selectedTab;
		if(page)
			page.realBounds = displayRectangle;
	}
	
	
	final bool multiline() // getter
	{
		return (_style() & TCS_MULTILINE) != 0;
	}
	
	
	final int rowCount() // getter
	{
		if(!created || !multiline)
			return 0;
		Message m;
		m = Message(hwnd, TCM_GETROWCOUNT, 0, 0);
		prevWndProc(m);
		return cast(int)m.result;
	}
	
	
	final void selectedIndex(int i) // setter
	{
		if(!created || !tchildren._pages.length)
			return;
		SendMessageA(hwnd, TCM_SETCURSEL, cast(WPARAM)i, 0);
	}
	
	
	// Returns -1 if there are no tabs selected.
	final int selectedIndex() // getter
	{
		if(!created || !tchildren._pages.length)
			return -1;
		Message m;
		m = Message(hwnd, TCM_GETCURSEL, 0, 0);
		prevWndProc(m);
		return cast(int)m.result;
	}
	
	
	final void selectedTab(TabPage page) // setter
	{
		int i;
		i = tabPages.indexOf(page);
		if(-1 != i)
			selectedIndex = i;
	}
	
	
	final TabPage selectedTab() // getter
	{
		int i;
		i = selectedIndex;
		if(-1 == i)
			return null;
		return tchildren._pages[i];
	}
	
	
	/+
	final void showToolTips(bool byes) // setter
	{
		if(byes)
			_style(_style() | TCS_TOOLTIPS);
		else
			_style(_style() & ~TCS_TOOLTIPS);
	}
	
	
	final bool showToolTips() // getter
	{
		return (_style() & TCS_TOOLTIPS) != 0;
	}
	+/
	
	
	final int tabCount() // getter
	{
		return tchildren._pages.length;
	}
	
	
	Rect getTabRect(int i)
	{
		Rect result;
		
		if(created)
		{
			RECT rt;
			Message m;
			m = Message(hwnd, TCM_GETITEMRECT, cast(WPARAM)i, cast(LPARAM)&rt);
			prevWndProc(m);
			if(!m.result)
				goto rtfail;
			result = Rect(&rt);
		}
		else
		{
			rtfail:
			with(result)
			{
				x = 0;
				y = 0;
				width = 0;
				height = 0;
			}
		}
		
		return result;
	}
	
	
	// drawItem event.
	EventHandler selectedIndexChanged;
	CancelEventHandler selectedIndexChanging;
	
	
	protected override void createParams(inout CreateParams cp)
	{
		super.createParams(cp);
		
		cp.className = TABCONTROL_CLASSNAME;
	}
	
	
	protected void onSelectedIndexChanged(EventArgs ea)
	{
		selectedIndexChanged(this, ea);
	}
	
	
	protected void onSelectedIndexChanging(CancelEventArgs ea)
	{
		selectedIndexChanging(this, ea);
	}
	
	
	protected override void onHandleCreated(EventArgs ea)
	{
		super.onHandleCreated(ea);
		
		SendMessageA(hwnd, TCM_SETPADDING, 0, MAKELPARAM(_pad.x, _pad.y));
		
		tchildren.doPages();
		
		// Bring selected tab to front.
		if(tchildren._pages.length)
		{
			int i;
			i = selectedIndex;
			if(-1 != i)
				tabToFront(tchildren._pages[i]);
		}
	}
	
	
	protected override void prevWndProc(inout Message msg)
	{
		msg.result = CallWindowProcA(tabcontrolPrevWndProc, msg.hWnd, msg.msg, msg.wParam, msg.lParam);
	}
	
	
	protected override void wndProc(inout Message m)
	{
		// TODO: support the tab control messages.
		
		switch(m.msg)
		{
			case WM_WINDOWPOSCHANGED:
				{
					WINDOWPOS* wp;
					wp = cast(WINDOWPOS*)m.lParam;
					
					if(!(wp.flags & SWP_NOSIZE) || (wp.flags & SWP_FRAMECHANGED))
					{
						if(tchildren._pages.length)
						{
							int i;
							i = selectedIndex;
							if(-1 != i)
							{
								tchildren._pages[i].realBounds = displayRectangle;
								//assert(tchildren._pages[i].bounds == displayRectangle);
							}
						}
					}
				}
				break;
			
			case WM_SETFOCUS:
				_exStyle(_exStyle() | WS_EX_CONTROLPARENT);
				break;
			
			case WM_KILLFOCUS:
				_exStyle(_exStyle() & ~WS_EX_CONTROLPARENT);
				break;
			
			case TCM_DELETEALLITEMS:
				m.result = FALSE;
				return;
			
			case TCM_DELETEITEM:
				m.result = FALSE;
				return;
			
			case TCM_INSERTITEMA:
			case TCM_INSERTITEMW:
				m.result = -1;
				return;
			
			//case TCM_REMOVEIMAGE:
			//	return;
			
			//case TCM_SETIMAGELIST:
			//	m.result = cast(LRESULT)null;
			//	return;
			
			case TCM_SETITEMA:
			case TCM_SETITEMW:
				m.result = FALSE;
				return;
			
			case TCM_SETITEMEXTRA:
				m.result = FALSE;
				return;
			
			case TCM_SETITEMSIZE:
				m.result = 0;
				return;
			
			case TCM_SETPADDING:
				return;
			
			case TCM_SETTOOLTIPS:
				return;
			
			default: ;
		}
		
		super.wndProc(m);
	}
	
	
	/+package+/ override LRESULT customMsg(inout CustomMsg msg) // package
	{
		TabPage page;
		NMHDR* nmh;
		nmh = cast(NMHDR*)msg.lParam;
		
		switch(nmh.code)
		{
			case TCN_SELCHANGE:
				page = selectedTab;
				if(page)
					tabToFront(page);
				
				onSelectedIndexChanged(EventArgs.empty);
				break;
			
			case TCN_SELCHANGING:
				{
					auto CancelEventArgs ea = new CancelEventArgs;
					onSelectedIndexChanging(ea);
					if(ea.cancel)
						return TRUE; // Prevent change.
					
					page = selectedTab;
					if(page)
						page.realVisible = false;
				}
				return FALSE; // Allow change.
			
			default: ;
		}
		
		return 0;
	}
	
	
	private:
	Point _pad = {x: 6, y: 3};
	TabPageCollection tchildren;
	
	
	void tabToFront(TabPage page)
	{
		page.realBounds = displayRectangle;
		SetWindowPos(page.handle, HWND_TOP, 0, 0, 0, 0, /+ SWP_NOACTIVATE | +/ SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW);
		assert(page.visible == true);
	}
	
	
	void updateTabText(TabPage page, char[] newText)
	in
	{
		assert(created);
	}
	body
	{
		int i;
		i = tabPages.indexOf(page);
		assert(-1 != i);
		
		TC_ITEMA tci;
		tci.mask = TCIF_TEXT;
		tci.pszText = toStringz(newText);
		Message m;
		m = Message(hwnd, TCM_SETITEMA, cast(WPARAM)i, cast(LPARAM)&tci);
		prevWndProc(m);
		
		// Updating a tab's text could cause tab rows to be adjusted,
		// so update the selected tab's area.
		page = selectedTab;
		if(page)
			page.realBounds = displayRectangle;
	}
}

