// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.timer;

private import dfl.winapi, dfl.event, dfl.base, dfl.application;


class Timer // docmain
{
	EventHandler tick;
	
	
	void enabled(bit on) // setter
	{
		if(on)
			start();
		else
			stop();
	}
	
	
	bit enabled() // getter
	{
		return timerId != 0;
	}
	
	
	final void interval(DWORD timeout) // setter
	{
		if(!timeout)
			throw new DflException("Invalid timer interval.");
		
		if(this.timeout != timeout)
		{
			this.timeout = timeout;
			
			if(timerId)
			{
				// I don't know if this is the correct behavior.
				// Reset the timer for the new timeout...
				stop();
				start();
			}
		}
	}
	
	
	final DWORD interval() // getter
	{
		return timeout;
	}
	
	
	final void start()
	{
		if(timerId)
			return;
		
		assert(timeout > 0);
		
		timerId = SetTimer(null, 0, timeout, &timerProc);
		if(!timerId)
			throw new DflException("Unable to start timer.");
		allTimers[timerId] = this;
	}
	
	
	final void stop()
	{
		if(timerId)
		{
			//delete allTimers[timerId];
			allTimers.remove(timerId);
			KillTimer(null, timerId);
			timerId = 0;
		}
	}
	
	
	~this()
	{
		dispose();
	}
	
	
	protected:
	
	void dispose()
	{
		stop();
	}
	
	
	void onTick(EventArgs ea)
	{
		tick(this, ea);
	}
	
	
	private:
	DWORD timeout = 100;
	UINT timerId = 0;
}


private:

Timer[UINT] allTimers;


extern(Windows) void timerProc(HWND hwnd, UINT uMsg, UINT idEvent, DWORD dwTime)
{
	try
	{
		if(idEvent in allTimers)
		{
			allTimers[idEvent].onTick(EventArgs.empty);
		}
		else
		{
			debug(APP_PRINT)
				printf("Unknown timer 0x%X.\n", idEvent);
		}
	}
	catch(Object e)
	{
		Application.onThreadException(e);
	}
}

