/*
	Copyright (C) 2005 Christopher E. Miller
	
	This software is provided 'as-is', without any express or implied
	warranty.  In no event will the authors be held liable for any damages
	arising from the use of this software.
	
	Permission is granted to anyone to use this software for any purpose,
	including commercial applications, and to alter it and redistribute it
	freely, subject to the following restrictions:
	
	1. The origin of this software must not be misrepresented; you must not
	   claim that you wrote the original software. If you use this software
	   in a product, an acknowledgment in the product documentation would be
	   appreciated but is not required.
	2. Altered source versions must be plainly marked as such, and must not be
	   misrepresented as being the original software.
	3. This notice may not be removed or altered from any source distribution.
*/


private import std.stdio, std.string, std.path, std.file;
private import std.process;
private import std.c.stdlib;

private import dfl.all, dfl.winapi;


private extern(Windows)
{
	DWORD GetLogicalDriveStringsA(DWORD nBufferLength,LPSTR lpBuffer);
	UINT GetDriveTypeA(LPCTSTR lpRootPathName);
	DWORD GetShortPathNameA(LPCSTR lpszLongPath, LPSTR lpszShortPath, DWORD cchBuffer);
	
	
	enum: UINT
	{
		DRIVE_FIXED = 3,
	}
}


enum Flags: DWORD
{
	NONE = 0,
	
	INSTALLED = 1, // Everything is setup.
}


RegistryKey rkey;
Flags flags = Flags.NONE;
char[] startpath, basepath;
char[] dmdpath;
bit isPrepared = false;

char[] optExet = "nt"; // Exe type.
char[] optSu = "console:4.0"; // Subsystem.
bit optForceInstall = false;


bit isValidDmdDir(char[] dir)
{
	char[] libdir;
	libdir = std.path.join(dir, "lib");
	if(std.path.isabs(dir) && std.file.exists(std.path.join(dir, "bin\\dmd.exe")) &&
		std.file.exists(libdir) && isdir(libdir))
		return true;
	return false;
}


void install()
{
	char[] s;
	
	/+
	switch(msgBox("Would you like to install DFL now?",
		"DFL", MsgBoxButtons.YES_NO, MsgBoxIcon.QUESTION))
	{
		case DialogResult.YES:
			break;
		
		default:
			exit(0);
	}
	+/
	
	if(dmdpath.length)
	{
		assert(isValidDmdDir(dmdpath));
		rkey.setValue("dmdpath", dmdpath);
	}
	else
	{
		writefln("Locating DMD...");
		
		char[128] drives;
		if(GetLogicalDriveStringsA(drives.length, drives))
		{
			char* p;
			for(p = drives; *p; p++)
			{
				if(GetDriveTypeA(p) == DRIVE_FIXED) // Only check fixed disks.
				{
					char[] s;
					s = std.path.join(.toString(p), "dmd");
					if(std.file.exists(s))
					{
						if(isValidDmdDir(s))
						{
							switch(msgBox("Found DMD at '" ~ s ~ "'.\r\n"
								"Would you like to use this path?\r\n\r\n"
								"Press No to keep looking.\r\n"
								"Press Cancel to abort and try again later.",
								"DFL", MsgBoxButtons.YES_NO_CANCEL, MsgBoxIcon.QUESTION))
							{
								case DialogResult.YES:
									rkey.setValue("dmdpath", s);
									goto found_dmd;
								
								case DialogResult.NO:
									break;
								
								default:
									exit(0);
							}
						}
						else
						{
							writefln("Found '%s' but no bin and lib directories...", s);
						}
					}
				}
				
				for(; *p; p++)
				{
				}
			}
		}
		
		// Didn't find DMD yet, so ask where it is.
		FolderBrowserDialog fbd;
		fbd = new typeof(fbd);
		fbd.description = "Please locate DMD.";
		browse_dmd_again:
		if(fbd.showDialog() != DialogResult.OK)
			exit(0); // Aborted.
		if(!isValidDmdDir(fbd.selectedPath))
		{
			fbd.description = "DMD was not found at that location. Please try again.";
			goto browse_dmd_again;
		}
		rkey.setValue("dmdpath", fbd.selectedPath);
	}
	
	found_dmd:
	
	flags |= Flags.INSTALLED;
	rkey.setValue("flags", cast(DWORD)flags);
	
	writef("Installation complete.\r\n\r\n");
}


void prepare()
{
	if(isPrepared)
		return;
	isPrepared = true;
	
	RegistryValueDword regDword;
	RegistryValueSz regSz;
	
	if(isValidDmdDir(basepath))
		dmdpath = basepath;
	
	regDword = cast(RegistryValueDword)rkey.getValue("flags");
	if(regDword)
		flags = cast(Flags)regDword.value;
	
	if(optForceInstall || !(flags & Flags.INSTALLED))
		install();
	
	
	void badInstall()
	{
		writefln("Bad install. To reinstall, use   dfl -dfl-i");
		exit(5);
	}
	
	
	if(!dmdpath.length)
	{
		regSz = cast(RegistryValueSz)rkey.getValue("dmdpath");
		if(regSz && isValidDmdDir(regSz.value))
			dmdpath = regSz.value;
		else
			badInstall();
	}
}


// Returns true if it's actually a DFL switch.
// The "-dfl-" part must be stripped.
bit doDflSwitch(char[] arg)
{
	int i;
	char[] equ = null;
	
	i = std.string.find(arg, '=');
	if(i != -1)
	{
		equ = arg[i + 1 .. arg.length];
		arg = arg[0 .. i];
	}
	
	
	void oops(char[] equName = "value")
	{
		writefln("Expected %s=<%s>", arg, equName);
		exit(2);
	}
	
	
	switch(arg)
	{
		case "dmd":
			prepare();
			std.process.system(quotearg(std.path.join(dmdpath, "bin\\dmd.exe\"")));
			exit(0);
		
		case "gui", "winexe":
			i = std.string.find(optSu, ':');
			optSu = "windows:" ~ optSu[i + 1 .. optSu.length];
			break;
		
		case "con", "console", "exe":
			i = std.string.find(optSu, ':');
			optSu = "console:" ~ optSu[i + 1 .. optSu.length];
			break;
		
		case "i":
			optForceInstall = true;
			break;
		
		/+
		case "h", "help", "?":
			showUsage();
			exit(0);
			break;
		+/
		
		case "exet", "exetype":
			if(equ.length)
			{
				optExet = equ;
			}
			else
			{
				oops();
			}
			break;
		
		case "su", "subsystem":
			if(equ.length > 3 && std.string.find(equ, ':') != -1)
			{
				optSu = equ;
			}
			else
			{
				oops("name:version");
			}
			break;
		
		case "doc":
			arg = std.path.join(basepath, "packages\\dfl\\doc\\index.html");
			if(!std.file.exists(arg))
				throw new Exception("'" ~ arg ~ "' not found");
			ShellExecuteA(null, null, std.string.toStringz(quotearg(arg)), null, null, 0);
			exit(0);
		
		case "readme":
			arg = std.path.join(basepath, "packages\\dfl\\readme.txt");
			if(!std.file.exists(arg))
				throw new Exception("'" ~ arg ~ "' not found");
			ShellExecuteA(null, null, std.string.toStringz(quotearg(arg)), null, null, 0);
			exit(0);
		
		case "tips":
			arg = std.path.join(basepath, "packages\\dfl\\tips.txt");
			if(!std.file.exists(arg))
				throw new Exception("'" ~ arg ~ "' not found");
			ShellExecuteA(null, null, std.string.toStringz(quotearg(arg)), null, null, 0);
			exit(0);
		
		case "examples", "samples", "eg", "ex":
			arg = std.path.join(basepath, "packages\\dfl\\examples");
			if(!std.file.exists(arg))
				throw new Exception("'" ~ arg ~ "' not found");
			ShellExecuteA(null, "explore", std.string.toStringz(quotearg(arg)), null, null, 0);
			exit(0);
		
		default:
			return false;
	}
	
	return true;
}


void showUsage()
{
	writefln("DFL written by Christopher E. Miller");
	writef("Usage:\n"
		"   dfl [<switches...>] <files...>\n\n");
	writef("Switches:\n"
		"   -dmd             Show DMD's usage.\n"
		"   -dfl-readme      Open the DFL readme.txt file.\n"
		"   -dfl-doc         Open the DFL documentation.\n"
		"   -dfl-tips        Open the DFL tips.txt file.\n"
		"   -dfl-eg          Explore the DFL examples directory.\n"
		"   -dfl-gui         Make a Windows GUI exe without a console.\n"
		"   -dfl-con         Make a console exe (default).\n"
		"   -dfl-exet=<x>    Override executable type.\n"
		"   -dfl-su=<x1:x2>  Override subsystem name and version.\n"
		"   -dfl-i           Force install.\n"
		"   <other>          Any other non-dfl switches are passed to DMD.\n");
	writef("Files:\n"
		"   Files passed to DMD.\n");
}


char[] quotearg(char[] fn)
{
	if(std.string.find(fn, ' ') != -1)
		return `"` ~ fn ~ `"`;
	return fn;
}


char[] getshortpath(char[] fn)
{
	char[] result;
	DWORD len;
	result = new char[MAX_PATH];
	
	len = GetShortPathNameA(toStringz(fn), result, result.length);
	return result[0 .. len];
}


char[] getParentDir(char[] dir)
{
	int i;
	
	for(;;)
	{
		if(!dir.length)
			return null;
		if(dir[dir.length - 1] == '/' || dir[dir.length - 1] == '\\')
			dir = dir[0 .. dir.length - 1];
		else
			break;
	}
	char[] result = null;
	for(i = dir.length - 1;;)
	{
		if(dir[i] == '/' || dir[i] == '\\')
		{
			result = dir[0 .. i];
			break;
		}
		if(!--i)
			break;
	}
	return result;
}


int main(/+ char[][] args +/)
{
	startpath = getshortpath(Application.startupPath);
	basepath = getParentDir(startpath);
	rkey = Registry.currentUser.createSubKey("Software\\DFL");
	
	char[][] dmdargs = null;
	int i;
	
	char[][] args;
	args = Environment.getCommandLineArgs();
	
	if(args.length > 1)
	{
		foreach(char[] _origarg; args[1 .. args.length])
		{
			if(_origarg.length && (_origarg[0] == '-' || _origarg[0] == '/'))
			{
				char[] arg;
				arg = _origarg[1 .. _origarg.length];
				i = std.string.find(arg, '-');
				
				if(i == -1)
					goto regular_switch;
				
				switch(arg[0 .. i])
				{
					case "dfl":
						if(!doDflSwitch(arg[i + 1 .. arg.length]))
						{
							writefln("Unrecognized DFL switch '-%s'", arg);
							return 1;
						}
						break;
					
					case "dmd":
						dmdargs ~= "-" ~ quotearg(arg[i + 1 .. arg.length]);
						break;
					
					default: regular_switch:
						if(!doDflSwitch(arg))
							dmdargs ~= quotearg(_origarg);
				}
			}
			else
			{
				dmdargs ~= quotearg(_origarg);
			}
		}
		
		prepare();
		
		if(!dmdargs.length)
		{
			showUsage();
		}
		else
		{
			char[] dfllib;
			char[] importdir;
			
			dfllib = std.path.join(basepath, "lib\\dfl.lib");
			if(!std.file.exists(dfllib))
			{
				dfllib = std.path.join(dmdpath, "lib\\dfl.lib");
				if(!std.file.exists(dfllib))
					throw new Exception("dfl.lib not found");
			}
			
			importdir = std.path.join(basepath, "import");
			if(!std.file.exists(importdir) || !std.file.exists(std.path.join(importdir, "dfl")))
			{
				importdir = std.path.join(dmdpath, "import");
				if(!std.file.exists(importdir) || !std.file.exists(std.path.join(importdir, "dfl")))
				{
					importdir = std.path.join(dmdpath, "src");
					if(!std.file.exists(importdir) || !std.file.exists(std.path.join(importdir, "dfl")))
					{
						importdir = std.path.join(dmdpath, "src\\phobos");
						if(!std.file.exists(importdir) || !std.file.exists(std.path.join(importdir, "dfl")))
							throw new Exception("DFL import directory not found");
					}
				}
			}
			
			dmdargs ~= "-I" ~ getshortpath(importdir);
			dmdargs ~= "-L/exet:" ~ optExet ~ "/su:" ~ optSu;
			dmdargs ~= getshortpath(dfllib);
			
			// Call DMD.
			assert(dmdpath.length);
			char[] cmdline;
			int sc;
			cmdline = quotearg(std.path.join(dmdpath, "bin\\dmd.exe")) ~ " " ~ std.string.join(dmdargs, " ");
			writefln("%s", cmdline);
			sc = std.process.system(cmdline);
			if(sc)
				writef("\nReturned status code %d\n", sc);
		}
	}
	else
	{
		prepare();
		showUsage();
	}
	
	return 0;
}

